define(function () {
    /**
     * Installs the Ace editor into the given HTML element.
     *
     * WARN: As of Vinyl 2.4, the install function will get called when
     *       switching from the display state to the edit state. The uninstall
     *       function only gets called when leaving the page.
     *       In addition, if a panel is initial display state edit, install will be run twice,
     *       once for view and another follow up as we switch to edit.
     * @param {HTMLElement} holderElement - HTML element into which the Ace
     *        editor should be installed.
     * @param {any} context - Widget context, including the Vinyl view model.
     * @returns {void}
     */
    function install(holderElement, context) {
        uninstall(holderElement, context);
        
        // Destructure dependencies.
        var ko = context.loadedDependencies[0];
        var resizeSensor = context.loadedDependencies[1];
        var autoCompleteMvSql = context.loadedDependencies[3];

        // Ace
        var mode = context.getParameter("mode");
        var theme = context.getParameter("theme");
        var fontSize = context.getParameter("fontSize");
        // normalize value, 1 or "1"
        var singleLine = context.getParameter("singleLine") == "1";

        // For more information on these parameters,
        // check dependencies/autoComplete-mvSQL.js
        var autoCompleteTableId = context.getParameter("autoComplete.tableId");
        var autoCompleteMvSqlQueryId = context.getParameter("autoComplete.mvSqlQueryId");

        var aceGlobal = ace;

        // Style
        var position = context.getParameter("position");
        var height = context.getParameter("height");
        var width = context.getParameter("width");
        var top = context.getParameter("top");
        var bottom = context.getParameter("bottom");
        var left = context.getParameter("left");
        var right = context.getParameter("right");

        // View model
        var viewModel = context.viewModel;
        // Cell
        var cell = context.getCell();

        // Add a collection of properties to the view model. These will be
        // available in the view.
        viewModel.ace = {
            style: {
                position: position,
                height: height,
                width: width,
                top: top,
                bottom: bottom,
                left: left,
                right: right
            },
            singleLine: singleLine,
            // NOTE: This works because the editor is recreated every time the
            //       control switches from the view to the edit state.
            readOnly: !context.isEditState(),
            cssClass: ko.pureComputed(function() {
                var lineClass = viewModel.ace.singleLine ? "ace_single_line vinyl-controls-text form-control vinyl-value" : "ace_multi_line";
                var readClass =  viewModel.ace.readOnly ? "ace_readonly" : "";
                return lineClass + " " + readClass;
            }),
            autoComplete: {
                tableId: autoCompleteTableId,
                mvSqlQueryId: autoCompleteMvSqlQueryId
            }
        };


        ko.applyBindingsToDescendants(viewModel, holderElement);

        // Construct the editor, stashing a reference in the view model.
        var editor = viewModel.ace.editor = aceGlobal.edit(holderElement.firstChild, {
            readOnly : viewModel.ace.readOnly,
            // Helps with the visual distinction of expecting or not input
            highlightActiveLine: !viewModel.ace.readOnly,
            highlightGutterLine: !viewModel.ace.readOnly,
            // More similar to the other browers input cursor
            cursorStyle: "slim",
            mode: "ace/mode/" + mode,
            theme: "ace/theme/" + theme,
            fontSize: fontSize
        });

        if (singleLine) {

            editor.setOptions({
                autoScrollEditorIntoView: true,
                highlightActiveLine: false,
                printMargin: false,
                showGutter: false
            });

            // Intercept new lines and tab characters
            editor.session.doc.insert = function(position, text) {
                var value = text.replace(/[\r\n\t]+/g, " ");

                return this.insertMergedLines(position, [ value ]);
            };

            editor.renderer.setScrollMargin(3, 3);

            // make mouse position clipping nicer
            editor.renderer.screenToTextCoordinates = function(x, y) {
                var pos = this.pixelToScreenCoordinates(x, y);
                return this.session.screenToDocumentPosition(
                    Math.min(this.session.getScreenLength() - 1, Math.max(pos.row, 0)),
                    Math.max(pos.column, 0)
                );
            };

            // Remove hotkeys from editor so it can be picked up
            // by Vinyl or the browser

            // enter/control enter
            editor.commands.removeCommand("addLineAfter");
            editor.commands.removeCommand("addLineBefore");

            // tab/shift tab
            editor.commands.removeCommand("indent");
            editor.commands.removeCommand("outdent");

            // control up/down
            editor.commands.removeCommand("scrolldown");
            editor.commands.removeCommand("scrollup");

            // up/down
            editor.commands.removeCommand("golineup");
            editor.commands.removeCommand("golinedown");

            // prevents enter to reach ace
            editor.commands.bindKey("Enter|Shift-Enter", "null")
        }

        if (viewModel.ace.autoComplete.tableId || viewModel.ace.autoComplete.mvSqlQueryId) {
            // custom code for supported auto completers
            autoCompleteMvSql.init(aceGlobal, editor, viewModel.ace.autoComplete);

            editor.on("focus", function() {
                // Open auto complete on focusing the field
                editor.execCommand("startAutocomplete");
            });
        }

        editor.sensor = new resizeSensor(editor.container, function() {
            editor.resize();
        });

        // Ace will throw an error if passed null.
        var value = cell.value || "";
        editor.session.setValue(value);

        if (singleLine) {
            // For single lines, default to have it all selected
            editor.selection.selectAll();
        }

        var localChange = false;

        // When the cell value changes, update the ACE editor
        // Sometimes we run a destroy/install, but other times we retain the same instance
        cell.setChangeCallback('value', function (value) {
            // When the value changes, set a flag so the ACE editor doesn't think 
            // the user changes it
            localChange = true;
            editor.session.setValue(value || "");
            
            // At this point, context getParameter is not updated yet
            // Delay 50ms, then ask for the new mode and set it
            // Updates the ACE to work with the latest mode (css/json/html etc)
            setTimeout(function(){
                var newMode = context.getParameter("mode");
                editor.session.setMode("ace/mode/" + newMode);
            }, 50);
            
            
            localChange = false;
        });

        // When the value changes in the session, do two things.
        editor.session.on("change", function (delta) {
            if (localChange) {
                return;
            }
            // Run the view modal touch, which will let vinyl know this cell changed.
            // We don't want to wait for the blur event for this, this activates the save icon.
            viewModel.touch();
        });

        var blurFunction = function (event) {

            // Since the editor can't handle null, we have to check
            // if the only change was making it into an empty string
            // or else it will dirty the cell and prompt the user
            // about unsaved changes even if they didn't do anything
            // beyond click on and away from the editor.
            var currentValue = cell.value || "";
            var newValue = editor.session.getValue();

            if (currentValue !== newValue) {
                // Set the session value in the view model cell value
                cell.value = newValue;
            }
        };

        // Run this blur when our textarea has a uncaught change event
        // The only time this runs is when Vinyl kicks off a hotkey (control-s) save event.
        // Vinyl tries to run the change on the event.target (see: commitInputChange)
        var textArea = $(holderElement).find(".ace_text-input");
        textArea.on("change.vinyl-ace", blurFunction);

        // Ace alone doesn't fire blur event on every case (eg.
        // paste and click away), so textare can help catch those
        // extra cases.
        textArea.on("blur.vinyl-ace", blurFunction);

        // Update the event cell is response to the Ace editor's normal blur event.
        editor.on("blur.vinyl-ace", blurFunction);
    }

    /**
     * Uninstalls the Ace editor from the given HTML element.
     *
     * WARN: As of Vinyl 2.4, the install function will get called when
     *       switching from the display state to the edit state. The uninstall
     *       function only gets called when leaving the page.
     * @param {HTMLElement} holderElement - HTML element into which the Ace
     *        editor should be installed.
     * @param {any} context - Widget context, including the Vinyl view model.
     * @returns {void}
     */
    function uninstall(holderElement, context) {
        var ace = context.viewModel.ace;
        if (ace) {
            var editor = ace.editor;

            // NOTE: Wrapping in a try/catch to work around this (currently)
            //       open issue:
            //
            //         https://github.com/marcj/css-element-queries/issues/72
            //
            //       Otherwise, an error occurs when attempting to dismiss error
            //       messages rendered at the bottom of the screen.
            //
            //       Once the issue has been fixed and the library updated, we
            //       can remove the try/catch.
            try { editor.sensor.detach(); } catch (e) { }

            editor.off("focus");
            editor.off("blur.vinyl-ace");
            editor.destroy();
            editor.container.remove();
            context.viewModel.ace = undefined;

            var textArea = $(holderElement).find(".ace_text-input");
            textArea.off("change.vinyl-ace");
            textArea.off("blur.vinyl-ace");
        }

        // NOTE: Wrapping in a try/catch to work around this (currently)
        //       open issue:
        //
        //         https://github.com/marcj/css-element-queries/issues/72
        //
        //       Otherwise, an error occurs when attempting to dismiss error
        //       messages rendered at the bottom of the screen.
        //
        //       Once the issue has been fixed and the library updated, we
        //       can remove the try/catch.
        try { sensor.detach(); } catch (e) { }
    }

    return {
        callbacks: {
            events: {
                install: install,
                uninstall: uninstall
            }
        }
    };
});