/* 
 * Supporting two different scopes for auto-complete
 * 1. Adding/Editing a columnn in BO (by providing autoCompleteOptions.tableId)
 *      list columns and functions, doesn't change during the lifetime of the 
 *      editor (so it builds a list once)
 * 2. mvSQL editor (not fully supported yet)
 *      list columns set changes (needs to be recalculated as tables are used,
 *      and aliases are set for both tables and columns);
 *      Has a fixed list of functions, linked data sources and tables associtated
 *      to them.   
 */

define(function () {

    var regexName = /^[a-zA-Z_][a-zA-Z0-9_]*$/;
    /**
     * For a given value, escapes it if needed it (if its not a simple identifier)
     * @param {string} value Value to be escaped, if needed.
     */
    function escapeName(value) {
        if (regexName.test(value)) {
            return value;
        }

        value = value.replace('"', '""');
        return '"' + value + '"';
    }

    /**
     * Load data from REST endpoint, applying a transformation
     * @param {string} url REST endpoint
     * @param {*} data data to be sent to the REST endpoint
     * @param {Function} transformResponse Format REST items into Ace's auto 
     * complete objects
     * @param {Function} handleSuccess calls with a list of ACE's auto 
     * complete objects once the request returns and processes successully
     */
    function load(url, data, transformResponse, handleSuccess) {

        $.get(url, data, function (response) {

            if (!response || response.status !== 200 || !response.items) {
                // we encontered some issue, fail silently as 
                // its not vital for the user (they should be
                // able to continue using the field) and ace 
                // will try again when more characters comes in
                return;
            }

            var autoComplete = [];

            for (var index = 0; index < response.items.length; index++) {
                var item = transformResponse(response.items[index]);
                autoComplete.push(item);
            }

            handleSuccess(autoComplete);
        });
    }

    /**
     * Load all rows (limiting to 10000)
     * @param {string} url REST endpoint
     * @param {*} data data to be sent to the REST endpoint
     * @param {Deferred} deferred Deferred object, expecting
     * a final list to send to ACE
     * @param {Function} transformResponse Format REST items into Ace's auto 
     * complete objects
     */
    function loadAll(url, data, deferred, transformResponse) {
        // fail safe
        var limitTotal = 10000;

        data.$limit = 100;
        data.$offset = 0;

        var fullAutoComplete = []; 

        var handleSuccess = function(autoComplete) { 

            if (fullAutoComplete.length) {
                // merge the two lists
                Array.prototype.push.apply(fullAutoComplete, autoComplete);
            } else {
                // it was empty, faster to just assign
                fullAutoComplete = autoComplete;
            }

            if(autoComplete.length >= data.$limit && 
                data.$limit + data.$offset < limitTotal) {

                data.$offset += data.$limit;
                load(url, data, transformResponse, handleSuccess);

            } else {

                deferred.resolve(fullAutoComplete);
            }
        };

        load(url, data, transformResponse, handleSuccess);
    }

    /**
     * Return all columns available to a tableId
     * @param {string} tableId TableId to check for used tables
     * @param {Deferred} deferred handle final list and send it 
     * to ACE
     */
    function loadColumns(tableId, deferred) {
        var url = "rest/v1/vinyl/column_autocompletelist";

        loadAll(url, {
            tableId: tableId,
        }, deferred, function(element) {
            var name;
            var value;

            if (element.tableNameOrAlias !== null) {
                value = escapeName(element.tableNameOrAlias) + "." + escapeName(element.name);
                name = element.tableNameOrAlias + "." + element.name;
            } else {
                value = escapeName(element.name);
                name = element.name;
            }

            return {
                caption: name,
                value: value,
                // Ace will overwrite the score depending on first character and
                // such, but if all else the same, columns will come before 
                // functions
                score: 10,
                meta: "Column"
            };
        });
    }

    /**
     * Return all functions available, if provided with tableId
     * the REST endpoint will check if its an union, and return
     * an empty list (we only support column names with a union)
     * @param {string} tableId TableId to check for unions
     * @param {Deferred} deferred handle final list and send it 
     * to ACE
     */
    function loadFunctions(tableId, deferred) {
        var url = "rest/v1/vinyl/function_autocompletelist";

        if (tableId) {
            url = "rest/v1/vinyl/function_autocompletelistfortable";
        }

        loadAll(url, {
            tableId: tableId
        }, deferred, function(element) {

            if (element.name === "Null")
            {
                // Special case, because Null and Null()
                // are both valid. So will default to
                // show the more common use.
                
                return {
                    caption: "null",
                    value: "null",
                    score: 1,
                    meta: "keyword"
                };
            }

            var argumentsText = "";
            var argCount = element.arguments;

            if (element.parameterSet && element.hasDefaultParameter) {
                // Expects parameters, but it has a default, so
                // don't bother the user. The only example currently
                // is:
                // now(LOCAL) vs now()
                argCount -= 1;
            }

            // turn the arguments into a template, eg.
            // DATEPART(${1:argument}, ${2:argument})
            for (var index = 0; index < argCount; index++) {
                argumentsText += (index === 0 ? "" : ", ") +
                                "${" + (index + 1) + ":argument}";
            }

            // Functions are set as snippets, they work as a template
            // that walks the user to set each argument
            return {
                caption: element.name,
                snippet: escapeName(element.name) + "(" + argumentsText + ")",
                type: "snippet",
                score: 1,
                meta: "Function"
            };
        });
    }

    /** 
     * Create the completers needed for Vinyl. As mentioned above, when possible
     * keeping a list of all items to avoid unecessary calls to the server
     * eg. function list don't change during the lifetime of a column edit/add
     */
    function setCompleters(languageTools, autoCompleteOptions) {

        var columnCompleter = {
            // Tells ace that " and . are part of a identifier ("Table"."Column")
            identifierRegexps: [/[a-zA-Z_0-9\".\$\-\u00A2-\uFFFF]/],
            getCompletions: function (editor, session, pos, prefix, callback) {

                if (!autoCompleteOptions.resolveColumns) {
                    var deferred = $.Deferred();
                    autoCompleteOptions.resolveColumns = $.when(deferred);
                    
                    loadColumns(autoCompleteOptions.tableId, deferred);
                }

                autoCompleteOptions.resolveColumns.then(function(columnAutoComplete) {
                    // callback(err, results)
                    callback(null, columnAutoComplete);
                });
            }
        };

        var functionCompleter = {
            getCompletions: function (editor, session, pos, prefix, callback) {
                
                if (!autoCompleteOptions.resolveFunctions) {
                    var deferred = $.Deferred();
                    autoCompleteOptions.resolveFunctions = $.when(deferred);
                    
                    loadFunctions(autoCompleteOptions.tableId, deferred);
                }

                autoCompleteOptions.resolveFunctions.then(function(functionsAutoComplete) {
                    // callback(err, results)
                    callback(null, functionsAutoComplete);
                });
            }
        };

        var completers = null;
        if (autoCompleteOptions.tableId) {
            // For columns, we only want our provided completer list
            completers = [ columnCompleter, functionCompleter ];

        } else {
            
            // DESIGNER NOTE:
            // Disabling all auto complete until V2, its more of an annuince until
            // auto complete can be context aware

            // For mvSQL, we want snippets, kerwords, and our given list of tables, functions
            completers = [ 
                // v2: [Linked.]tableName and columns, with context/scope
                // functionCompleter, // all functions
                // languageTools.keyWordCompleter, // provided all keywords (select, from, where, ...)
                // languageTools.snippetCompleter // provided mode snippets (select ...)
            ];
        }

        // Overwrite acer's
        languageTools.setCompleters(completers);
    };


    // The symbols we will ignore to open auto complete on
    var SYMBOLS_REGEX = /[|+\-*/<>=]/;

    /**
     * Ace's default live auto complete always wait to have some prefix
     * and also don't apply when user is deleting content. 
     * This approach is more aggresive, only not opening it if the user
     * just reached/typed a symbol (+ - / * < > =), as usually you would
     * want a blank space first.
     * @param {*} e ace afterExec event object
     */
    function doCustomLiveAutocompletion(e) {
        var editor = e.editor;
        var hasCompleter = editor.completer && editor.completer.activated;
    
        if (!hasCompleter 
            && (e.command.name === "insertstring" || e.command.name === "backspace")) {
            
            var pos = editor.getCursorPosition();
            var line = editor.session.getLine(pos.row);
            var isSymbol = pos.column > 0 && SYMBOLS_REGEX.test(line[pos.column-1]);

            if (!isSymbol) {
                var completer = autoComplete.for(editor);
                // Disable autoInsert
                completer.autoInsert = false;
                completer.showPopup(editor);
            }
        }
    };

    /**
     * The auto complete pop up don't rezise to fit its
     * content out of the box, so when its available
     * add that logic to its renderer.
     */
    function setUpdatePopUpSize(editor){
        
        if (!editor.completer) {
            return;
        }

        var popup = editor.completer.getPopup();

        if (!editor.completer.popup || popup.updateSize) {
            return;
        }

        var renderer = popup.renderer;

        popup.updateSize = function updateSize(e, renderer) {
            var maxWidth = editor.renderer.$size.width;
            for (var index = 0; index < renderer.session.doc.getLength(); index++)
            {
                var text = renderer.session.getLine(index);
                var chars = renderer.session.$getStringScreenWidth(text)[0];

                var width = Math.max(chars, 2) * renderer.characterWidth // text size
                            + 2 * renderer.$padding // padding
                            + 2 // little extra for the cursor
                            + 0 // add border width if needed
                            + 75; // buffer for the meta data

                maxWidth = Math.max(maxWidth, width);
            }

            var height = renderer.$size.height;
            if (height <= 0) 
            {
                    // Zoom bug, ace will sometimes not set a height
                    // if some conditions are met (one of them being
                    // zoomed out 70/75%).
                    height = 168;
            }

            // update container size
            renderer.container.style.width = maxWidth + "px";
            renderer.container.style.height = height + "px";
            // update computed size stored by the editor
            renderer.onResize(false, 0, maxWidth, height);
        }

        renderer.on("beforeRender", popup.updateSize);
        popup.updateSize(null, renderer);
    }
    

    var languageTools;
    var autoComplete;
    function init(aceGlobal, editor, autoCompleteOptions) {

        aceGlobal.config.loadModule("ace/ext/language_tools", function(tools) {
            languageTools = tools;
            autoComplete = aceGlobal.require("ace/autocomplete").Autocomplete;

            setCompleters(languageTools, autoCompleteOptions);

            // Add some extra logic when creating the completer for an editor
            // 1. Inserting a value keeps the cursor in view
            // 2. Updates cell value
            // 3. Remove tab hotkey as an option selector
            if (!autoComplete._originalFor) {

                autoComplete._originalFor = autoComplete.for;
    
                autoComplete.for = function(aceEditor) {
                    var completer = autoComplete._originalFor(aceEditor);
        
                    // Auto complete don't behave quite well on keeping
                    // the cursor in view, which on a single line context
                    // leads to a very poor UI experience. So wait until
                    // the completer is available to intercept inserting
                    // any match (snippet or otherwise).
        
                    if (!completer._originalInsertMatch) {
                        completer._originalInsertMatch = completer.insertMatch;
                        completer.insertMatch = function(data, options) {

                            // If the user picks an item without ever typing into the
                            // field, ace fails to recgonize it was focused
                            aceEditor.focus();

                            // Insert new value
                            this._originalInsertMatch(data, options);

                            // Makes sure the cursor is in view
                            aceEditor.renderer.scrollCursorIntoView();
                        };
                    }
    
        
                    // Stop tab from selecting auto complete (enter will still
                    // do it)
                    completer.keyboardHandler.removeCommand("Tab");
        
                    return completer;
                };
            }
                    
            editor.commands.on("afterExec", function(e) {
                // enable custom live auto complete
                doCustomLiveAutocompletion(e);
                // check if popup is available and resizing
                setUpdatePopUpSize(e.editor);
            });

            editor.setOptions({
                enableBasicAutocompletion: true,  // crtl+spacebar
                enableSnippets: true // templates
            });
        });
    }

    return {
        init: init
    };

});